/*
 * Connections.cpp
 *
 *  Created on: 10/08/2009
 *      Author: rcamargo
 */

#include "Connections.hpp"
//#include "SynapticChannels.hpp"
#include <cstdlib>
#include <cstdio>
#include <cassert>

Connections::Connections() {
}

Connections::~Connections() {
	// TODO: Remove connections
}

std::vector<Conn *> & Connections::getConnArray (int source) {
	return connMap[source];
}

int Connections::createTestConnections () {

	int source;

	Conn *conn1 = new Conn();
	source = CONN_NEURON_TYPE*0 + 0;
	conn1->dest   = CONN_NEURON_TYPE*3 + 1;
	conn1->synapse = 0;
	conn1->weigth = 1;
	conn1->delay = 10;
	connMap[source].push_back(conn1);

	Conn *conn2 = new Conn();
	source = CONN_NEURON_TYPE*0 + 0;
	conn2->dest   = CONN_NEURON_TYPE*1 + 1;
	conn2->synapse = 0;
	conn2->weigth = 1;
	conn2->delay = 10;
	connMap[source].push_back(conn2);

	Conn *conn3 = new Conn();
	source = CONN_NEURON_TYPE*3 + 1;
	conn3->dest   = CONN_NEURON_TYPE*1 + 0;
	conn3->synapse = 1;
	conn3->weigth = 1;
	conn3->delay = 10;
	connMap[source].push_back(conn3);

	return 0;
}

/**
 * type [0][1][2][3] pyramidal
 * type [4][5][6][7] inhibitory
 */
int Connections::connectAssociativeFromFile (char *filename) {
	return 0;
}

int Connections::connectRandom (ftype pyrRatio, ftype inhRatio, int *typeList, int nTypes, int *nNeurons, SharedNeuronGpuData *sharedData, int threadNumber) {

	int nConnTotal = 0;

	int nPyramidal  = 0;
	int nInhibitory = 0;
	for (int type=0; type < nTypes; type++) {
		if (typeList[type] == PYRAMIDAL_CELL)
			nPyramidal += nNeurons[type];
		else if (typeList[type] == INHIBITORY_CELL)
			nInhibitory += nNeurons[type];
	}

	/**
	 * Connects the pyramidal-pyramidal cells
	 */
	for (int sType=0; sType < nTypes; sType++) {

		if (typeList[sType] == INHIBITORY_CELL) continue;

		for (int sNeuron=0; sNeuron < nNeurons[sType]; sNeuron++) {

			/**
			 * Connects to a random neuron k times
			 */
			for (int c=0; c < nPyramidal * pyrRatio; c++) {

				Conn *conn1 = new Conn();
				conn1->synapse = 0;

				int32_t wei;
				random_r( sharedData->randBuf[threadNumber], &wei );
				conn1->weigth = sharedData->excWeight * (0.5 + ((ftype)wei)/RAND_MAX);
				//conn1->weigth = sharedData->excWeight;

				int32_t del;
				random_r( sharedData->randBuf[threadNumber], &del );
				conn1->delay = 10 + 10.*((ftype)del)/RAND_MAX;
				//conn1->delay = 10;

				int32_t pos;
				random_r( sharedData->randBuf[threadNumber], &pos );
				conn1->dest = pos % nPyramidal;

				int count = 0;
				for (int dType=0; dType < nTypes; dType++) {

					if (typeList[dType] == INHIBITORY_CELL) continue;

					if (conn1->dest < count + nNeurons[dType]) {
						conn1->dest = CONN_NEURON_TYPE*dType + (conn1->dest-count);
						break;
					}
					else
						count += nNeurons[dType];
				}

				connMap[CONN_NEURON_TYPE*sType + sNeuron].push_back(conn1);
				nConnTotal++;
			}
		}
	}

	/**
	 * Connects the pyramidal-inhibitory cells
	 */
	for (int sType=0; sType < nTypes; sType++) {

		if (typeList[sType] == INHIBITORY_CELL) continue;

		for (int sNeuron=0; sNeuron < nNeurons[sType]; sNeuron++) {

			/**
			 * Connects to a random neuron k times
			 */
			for (int c=0; c < nInhibitory * inhRatio; c++) {

				Conn *conn1 = new Conn();
				conn1->synapse = 0;

				int32_t wei;
				random_r( sharedData->randBuf[threadNumber], &wei );
				conn1->weigth = sharedData->pyrInhWeight * (0.5 + ((ftype)wei)/RAND_MAX);
				//conn1->weigth = sharedData->pyrInhWeight;

				int32_t del;
				random_r( sharedData->randBuf[threadNumber], &del );
				conn1->delay = 10 + ((ftype)del)/RAND_MAX;
				//conn1->delay = 10;

				int32_t pos;
				random_r( sharedData->randBuf[threadNumber], &pos );
				conn1->dest = pos % nInhibitory;

				int count = 0;

				for (int dType=0; dType < nTypes; dType++) {

					if (typeList[dType] == PYRAMIDAL_CELL) continue;

					if (conn1->dest < count + nNeurons[dType] ) {
						conn1->dest = CONN_NEURON_TYPE*dType + (0.5 + (conn1->dest-count));
						break;
					}
					else
						count += nNeurons[dType];
				}


				connMap[CONN_NEURON_TYPE*sType + sNeuron].push_back(conn1);
				nConnTotal++;
			}

			//printf("\n");
		}
	}

	/**
	 * Connects the inhibitory-pyramidal cells
	 */
	int inhNeuron = 0;
	if (inhRatio > 0) {
		for (int sType=0; sType < nTypes; sType++) {

			if (typeList[sType] == PYRAMIDAL_CELL) continue;

			for (int sNeuron=0; sNeuron < nNeurons[sType]; sNeuron++, inhNeuron++) {

				Conn *conn1 = new Conn();
				conn1->synapse = 1; // inhibitory

				//int32_t wei;
				//random_r( sharedData->randBuf[threadNumber], &wei );
				//conn1->weigth = sharedData->inhPyrWeight * ((ftype)wei)/RAND_MAX;
				conn1->weigth = sharedData->inhPyrWeight;

				int32_t del;
				random_r( sharedData->randBuf[threadNumber], &del );
				conn1->delay = 10 + ((ftype)del)/RAND_MAX;
				//conn1->delay = 10;

				conn1->dest = inhNeuron;
				int count = 0;
				for (int dType=0; dType < nTypes; dType++) {

					if (typeList[dType] == INHIBITORY_CELL) continue;

					if (conn1->dest < count + nNeurons[dType]) {
						conn1->dest = CONN_NEURON_TYPE*dType + (conn1->dest-count);
						break;
					}
					else
						count += nNeurons[dType];
				}


				connMap[CONN_NEURON_TYPE*sType + sNeuron].push_back(conn1);
				nConnTotal++;
			}
		}
	}

	printf("Total number of connections = %d.\n", nConnTotal);

	//assert(false);
	return 0;
}

