/*
 * HinesGpu.cu
 *
 *  Created on: 06/06/2009
 *      Author: rcamargo
 */

/**
 * TODO: otimizar
 * - acessar sTriangList com POS
 * - Ver acessos à memória global
 */

/**
 * TODO:
 * - Executar em múltiplas GPUs [OK]
 *
 * - Implementar neurônios de minha simulação
 * - Implementar rede com múltiplos neurônios
 *
 * - Testar desempenho e comparar com GENESIS
 *
 * - Otimizar (usar memória compartilhada)
 * - Otimizar (reduzir número de alocações de memória para copiar lista de spikes)
 */

//extern "C" {
#include "HinesMatrix.hpp"
#include "PlatformFunctions.hpp"
#include "HinesStruct.hpp"
#include <cassert>

//#define POS(i) (i) + nComp*threadIdx.x
#define POS(i) (i)*blockDim.x+threadIdx.x
//#define POS1(i) (i) + leftListSize*threadIdx.x
#define POS1(i) (i)*blockDim.x + threadIdx.x

#define NLOCALSPIKES 16 // 16

__device__ void printMatrix(HinesStruct *hList, ftype *list, ftype *rhsLocal, ftype *vmListLocal) {


	/*
	printf ("----------------------------------------------------------\n");
	int neuron = blockIdx.x * blockDim.x + threadIdx.x;
	HinesStruct & h = hList[neuron];
	int nComp = h.nComp;

	int pos = 0;
	ftype zero = 0;
	for (int i=0; i<nComp; i++) {
		for (int j=0; j<nComp; j++) {

			if (i == h.leftListLine[pos] && j == h.leftListColumn[pos])
				printf( "%10.4e\t", list[pos++]);
			else
				printf( "%10.4e\t", zero);
		}
		printf( "%10.2f\t", vmListLocal[POS(i)]); // mV
		printf( "%10.2e\t\n", rhsLocal[POS(i)]);
	}
	*/
}

/***************************************************************************
 * This part is executed in every integration step
 ***************************************************************************/

__device__ ftype getCurrent(HinesStruct *hList, ftype *vmListLocal,
		ucomp synType, ftype spikeTime, ftype weight, ucomp comp, ftype currTime,
		ftype *sTau, ftype *sGmax, ftype *sEsyn) {

	ftype gsyn = 0;
	ftype current = 0;
	if (synType == SYNAPSE_AMPA) {
		ftype r = (currTime - spikeTime) / sTau[2*SYNAPSE_AMPA];
		gsyn = sGmax[SYNAPSE_AMPA] * r * expf(1 - r) * weight;
		current = (vmListLocal[POS(comp)] - sEsyn[SYNAPSE_AMPA]) * gsyn;
	}

	else if (synType == SYNAPSE_GABA) {
		ftype r = (currTime - spikeTime) / sTau[2*SYNAPSE_GABA];
		gsyn = sGmax[SYNAPSE_GABA] * r * expf(1 - r) * weight;
		current = (vmListLocal[POS(comp)] - sEsyn[SYNAPSE_GABA]) * gsyn;
	}



	return current;
}

__device__ void findSynapticCurrents(HinesStruct *hList, ftype *active, ftype *vmListLocal,
		ftype currTime, ftype *sTau, ftype *sGmax, ftype *sEsyn) {

	//ftype *freeMem;

	HinesStruct & h = hList[blockIdx.x * blockDim.x + threadIdx.x];

	int synapseListSize = h.synapseListSize;
	int spikeListSize   = h.spikeListSize;

	ftype *spikeList = h.spikeList;
	ftype *synapseWeightList = h.synapseWeightList;

	if (spikeListSize > 0) {

		for (int syn=0, spike=0; syn < synapseListSize; syn++) {
			int synComp = h.synapseCompList[syn];
			int lastSpikePos = spikeListSize;
			if ( syn < (synapseListSize-1) )
				lastSpikePos = h.synSpikeListPos[syn+1];
			ucomp synapseType = h.synapseTypeList[syn];

			for (; spike < lastSpikePos; spike += 2 ) {

				ftype spk1 = spikeList[spike];
				ftype spk2 = ((spike+1) < lastSpikePos) ? spikeList[spike+1] : 1e100000;
				ftype w1 = synapseWeightList[spike];
				ftype w2 = ((spike+1) < lastSpikePos) ? synapseWeightList[spike+1] : 0;

				ftype current = 0;
				if (currTime >= spk1)
					current += getCurrent(hList, vmListLocal, synapseType, spk1, w1, synComp, currTime, sTau, sGmax, sEsyn);
				if (currTime >= spk2)
					current += getCurrent(hList, vmListLocal, synapseType, spk2, w2, synComp, currTime, sTau, sGmax, sEsyn);

				active [POS(synComp)] += current;
			}

			spike = lastSpikePos;
		}
	}

}

/**
 * Find the gate openings in the next time step
 * m(t + dt) = a + b m(t - dt)
 */
__device__ void evaluateGatesG( HinesStruct *hList, ftype *vmListLocal,
		ftype *nGate, ftype *hGate, ftype *mGate ) {

	//int neuron = blockIdx.x * blockDim.x + threadIdx.x;
	HinesStruct & h = hList[blockIdx.x * blockDim.x + threadIdx.x];

	ftype alpha, beta, a, b;
	ftype V;
	ftype dtRec = 1/h.dt;

	for (int i=0; i<h.compListSize; i++) {
		V = vmListLocal[ POS(h.compList[i]) ];

		// gate m
		alpha = (V != 25.0) ? (0.1 * (25 - V)) / ( expf( 0.1 * (25-V) ) - 1 ) : 1; // ____expff para double
		beta  = 4 * expf( -V/18 );
		a = alpha / (dtRec + (alpha + beta)/2);
		b = (dtRec - (alpha + beta)/2) / (dtRec + (alpha + beta)/2);
		mGate[POS(i)] = a + b * mGate[POS(i)];

		// gate h
		alpha =  0.07 * expf ( -V/20 );
		beta  = 1 / ( expf( (30-V)/10 ) + 1 );
		a = alpha / (dtRec + (alpha + beta)/2);
		b = (dtRec - (alpha + beta)/2) / (dtRec + (alpha + beta)/2);
		hGate[POS(i)] = a + b * hGate[POS(i)];

	 	// gate n
		alpha = (V != 10.0) ? (0.01 * (10 - V)) / ( expf( 0.1 * (10-V) ) - 1 ) : 0.1;
		beta  = 0.125 * expf ( -V/80 );
		a = alpha / (dtRec + (alpha + beta)/2);
		b = (dtRec - (alpha + beta)/2) / (dtRec + (alpha + beta)/2);
		nGate[POS(i)] = a + b * nGate[POS(i)];

	}
}

__device__ void findActiveCurrentsG(HinesStruct *hList, ftype *activeList, ftype *vmListLocal,
		ftype *nGate, ftype *hGate, ftype *mGate, int nComp) {

	evaluateGatesG(hList, vmListLocal, nGate, hGate, mGate);

	int neuron = blockIdx.x * blockDim.x + threadIdx.x;
	HinesStruct & h = hList[neuron];

	/**
	 * Update the channel conductances
	 */
	ftype Ek  = h.EK;
	ftype ENa = h.ENa;
	for (int i=0; i<h.compListSize; i++) {
		ftype gNaChannel = h.gNaBar[i] * mGate[POS(i)] * mGate[POS(i)] * mGate[POS(i)] * hGate[POS(i)];
		ftype gKChannel  =  h.gKBar[i] * nGate[POS(i)] * nGate[POS(i)] * nGate[POS(i)] * nGate[POS(i)];

		int comp = h.compList[i];
		ftype active = - gNaChannel * ENa - gKChannel  * Ek  ;
		active -=  ( 1 / h.Rm[comp] ) * ( h.ELeak );

		activeList[ POS(comp) ] += active;
		h.gNaChannel[i] = gNaChannel;
		h.gKChannel[i] = gKChannel;
	}

}

__device__ void upperTriangularizeAll(HinesStruct *hList, ftype *sTriangList,
				ftype *sLeftList, ucomp *sLeftListLine, ucomp *sLeftListColumn,
				ucomp *sLeftStartPos, ftype *rhsLocal, ftype *vmListLocal,
				ftype *nGate, ftype *hGate, ftype *mGate,
				ftype *sTau, ftype *sGmax, ftype *sEsyn,
				ftype *freeMem) {

	int neuron = blockIdx.x * blockDim.x + threadIdx.x;
	HinesStruct & h = hList[neuron];

	ftype *active = freeMem;
	ftype *Cm = h.Cm;
	ftype *curr = h.curr;

	int nComp = h.nComp;
	int leftListSize = h.leftListSize;

	for (int i=0; i<nComp; i++) {
		active[ POS(i)] = 0;
	}

	//__syncthreads();

	findActiveCurrentsG(hList, active, vmListLocal, nGate, hGate, mGate, nComp);

	findSynapticCurrents(hList, active, vmListLocal, h.currStep * h.dt, sTau, sGmax, sEsyn);


//	//int neuron = blockIdx.x * blockDim.x + threadIdx.x;
//	if (h.type == 0 && neuron == 1) {
//		for (int i=0; i<nComp; i++)
//			printf("active=%f\n", active[POS(i)]);
//		printf("\n");
//	}


	ftype dtRec = 1/h.dt;
	//rhsLocal[POS(0)] = (-2) * vmListLocal[POS(0)] * Cm[0] * dtRec - curr[0] + active[POS(0)];
	for (int i=0; i<nComp; i++)
		rhsLocal[POS(i)] = (-2) * vmListLocal[POS(i)] * Cm[i] * dtRec - curr[i] + active[POS(i)];

	// ***
	// 1000ms 960 16 1 -> 0.125ms
	for (int k = 0; k < leftListSize; k++)
		sTriangList[k] = sLeftList[k];

	for (int i = 0; i < h.compListSize; i++) {

		int comp = h.compList[i];
		int pos = sLeftStartPos[ comp ];

		for (; sLeftListColumn[pos] < comp && pos < leftListSize ; pos++);

		sTriangList[pos] -= (h.gNaChannel[i] + h.gKChannel[i]);
	}


	// 1000ms 960 16 1 -> 0.640ms
	for (int k = 0; k < leftListSize; k++) {

		int c = sLeftListColumn[k];
		int l = sLeftListLine[k];

		if( c < l ) {

			int pos = sLeftStartPos[c];
			for (; c == sLeftListLine[pos]; pos++)
				if (sLeftListColumn[pos] == c)
					break;

			double mul = -sTriangList[k] / sTriangList[pos];

			pos = sLeftStartPos[c];
			int tempK = sLeftStartPos[l];

			for (; c == sLeftListLine[pos] && pos < leftListSize; pos++) {
				for (; sLeftListColumn[tempK] < sLeftListColumn[pos] && tempK < leftListSize ; tempK++);

				sTriangList[tempK] += sTriangList[pos] * mul;
			}
			rhsLocal[POS(l)] += rhsLocal[POS(c)] * mul;
		}
	}


}



__device__ void updateRhsG(HinesStruct *hList, 
						   ftype *sMulList, ucomp *sMulListComp, ucomp *sMulListDest,
						   ftype *rhsLocal, ftype *vmListLocal,
						   ftype *nGate, ftype *hGate, ftype *mGate,
						   ftype *sTau, ftype *sGmax, ftype *sEsyn,
						   ftype *freeMem) {

	int neuron = blockIdx.x * blockDim.x + threadIdx.x;
	HinesStruct & h = hList[neuron];

	ftype *active = freeMem;
	ftype *Cm = h.Cm;
	ftype *curr = h.curr;

	int nComp = h.nComp;

	for (int i=0; i<nComp; i++) {
		active[ POS(i)] = 0;
	}

	//__syncthreads();

	findActiveCurrentsG(hList, active, vmListLocal, nGate, hGate, mGate, nComp);
	findSynapticCurrents(hList, active, vmListLocal, h.currStep * h.dt, sTau, sGmax, sEsyn);

	ftype dtRec = 1/h.dt;
	for (int i=0; i<nComp; i++)
		rhsLocal[POS(i)] = (-2) * vmListLocal[POS(i)] * Cm[i] * dtRec - curr[i] + active[POS(i)];

	int mulListSize = h.mulListSize;
	for (int mulListPos = 0; mulListPos < mulListSize; mulListPos++) {
		int dest = sMulListDest[mulListPos];
		int pos  = sMulListComp[mulListPos];
		rhsLocal[POS(dest)] += rhsLocal[POS(pos)] * sMulList[mulListPos];
	}

}

__device__ void backSubstituteG(HinesStruct *hList, 
								ftype *sTriangList, ucomp *sLeftListLine, ucomp *sLeftListColumn, 
								ftype *rhsLocal, ftype *vmListLocal, ftype* freeMem) {

	int neuron = blockIdx.x * blockDim.x + threadIdx.x;
	HinesStruct & h = hList[neuron];

	int nComp = h.nComp;
	//int index1;
	int leftListSize = h.leftListSize;

	ftype *vmTmpLocal = freeMem;

	//index1 = nComp-1;

	if (h.triangAll == 0 && h.compListSize > 0) // has active channels only in soma
		vmTmpLocal[POS(nComp-1)] = rhsLocal[POS(nComp-1)] / ( sTriangList[(leftListSize-1)] - h.gNaChannel[0] - h.gKChannel[0] );
	else
		vmTmpLocal[POS(nComp-1)] = rhsLocal[POS(nComp-1)] / sTriangList[(leftListSize-1)];


	ftype tmp = 0;
	for (int leftListPos = leftListSize-2; leftListPos >=0 ; leftListPos--) {
		int line   = sLeftListLine[(leftListPos)];
		int column = sLeftListColumn[(leftListPos)];
		if (line == column) {
			vmTmpLocal[POS(line)] = (rhsLocal[POS(line)] - tmp) * (1 / sTriangList[(leftListPos)]);
			tmp = 0;
		}
		else
			tmp += vmTmpLocal[POS(column)] * sTriangList[(leftListPos)];
	}

	for (int l = 0 ; l < nComp; l++)
		vmListLocal[POS(l)] = 2 * vmTmpLocal[POS(l)] - vmListLocal[POS(l)];

	//if (h.type == 0 && neuron == 1) printf("vmList=%.4f\n", vmListLocal[POS(0)]);

}

__global__ void solveMatrixG(HinesStruct *hList, int nSteps, int nNeurons, ftype *spikeListGlobal, ftype *weightListGlobal, int *spikeListPosGlobal, int *spikeListStartGlobal, ftype *vmListGlobal) {


	//printf("Starting Kernel...\n");

	int neuron = blockIdx.x * blockDim.x + threadIdx.x;
	if (neuron >= nNeurons) return;
	HinesStruct & h = hList[neuron];
	int nComp = h.nComp;
	int triangAll = h.triangAll;
	//int type = h.type;

	if (h.currStep > 0) {
		h.synSpikeListPos = spikeListPosGlobal + neuron * h.synapseListSize;
		h.spikeList = spikeListGlobal + spikeListStartGlobal[neuron];
		h.synapseWeightList = weightListGlobal + spikeListStartGlobal[neuron];
		h.spikeListSize = spikeListStartGlobal[neuron+1]-spikeListStartGlobal[neuron]; //spikeListPosGlobal[(neuron+1) * h.synapseListSize] - h.synSpikeListPos[0];
	}
	else {
		h.synSpikeListPos = 0;
		h.spikeList = 0;
		h.synapseWeightList = 0;
		h.spikeListSize = 0;
	}

	//	if (neuron==0)
	//		h.curr[0] = 10e-4;// + neuron*1e-4;

	// (ftype * 5 + ucomp * 10) * nComp
	// ftype=4 e ucomp=2 e ncomp = 8   ->  320 bytes 
	// ftype=4 e ucomp=2 e ncomp = 64  -> 2560 bytes  
	extern __shared__ ftype sharedMem[]; 
	ftype *sLeftList       = (ftype *)sharedMem;
	ucomp *sLeftListLine   = (ucomp *)&(sLeftList[h.leftListSize]); 	
	ucomp *sLeftListColumn = (ucomp *)&(sLeftListLine[h.leftListSize]); 

	ftype *sMulList     = (ftype *)&(sLeftListColumn[h.leftListSize]); // mulSize is zero when triangAll is 1
	ucomp *sMulListComp = (ucomp *)&(sMulList[h.mulListSize]); 	
	ucomp *sMulListDest = (ucomp *)&(sMulListComp[h.mulListSize]); 

	ucomp *sLeftStartPos = (ucomp *)&(sMulListDest[h.mulListSize]);


	int nChannelTypes = h.nChannelTypes;
	ftype *sTau		= (ftype *)&(sLeftStartPos[nComp]);
	ftype *sGmax 	= (ftype *)&(sTau[nChannelTypes*2]);
	ftype *sEsyn 	= (ftype *)&(sGmax[nChannelTypes]);

	ftype *lastSharedAddress = (ftype *)&(sEsyn[nChannelTypes]);

	for (int id=0; id < nChannelTypes; id++ ) {
		sTau[2*id]   = h.tau[2*id];
		sTau[2*id+1] = h.tau[2*id+1];
		sGmax[id] 	 = h.gmax[id];
		sEsyn[id] 	 = h.esyn[id];

	}

	for (int k=0; k < nComp; k ++ ) {
		sLeftStartPos[k] = h.leftStartPos[k];
	}

	for (int k=0; k < h.leftListSize; k ++ ) {
		if (triangAll == 0) sLeftList[k] = h.triangList[k];
		else				sLeftList[k] = h.leftList[k];
		sLeftListLine[k]   = h.leftListLine[k];
		sLeftListColumn[k] = h.leftListColumn[k];
	}

	if (triangAll == 0) {
		for (int k=0; k < h.mulListSize; k ++ ) {
			sMulList[k]     = h.mulList[k];
			sMulListComp[k] = h.mulListComp[k];
			sMulListDest[k] = h.mulListDest[k];
		}
	}


	//printf ("SolveMatrixG: Ok0\n");

	/* 
	 * Allocate for each individual neuron
	 * nThreads * nComp * ftype * 2
	 * 32 * [8 ] * 4 * 2 = 32 * 64 = 2K 	
	 * 32 * [32] * 4 * 2 =         = 8K
	 */
	ftype *rhsLocal = (ftype *)lastSharedAddress;
	//ftype *rhsLocal = (ftype *)&(sMulListDest[h.mulListSize]);
	ftype *vmListLocal = rhsLocal + blockDim.x * nComp;

	ftype *nGate = vmListLocal + blockDim.x * nComp;
	ftype *hGate = nGate + blockDim.x * h.compListSize;
	ftype *mGate = hGate + blockDim.x * h.compListSize;

	ftype *freeMem = mGate + blockDim.x * h.compListSize;
	ftype *sTriangList = 0;

	//printf ("used sharedMem1=%ld\n", (long)freeMem-(long)sharedMem);

	if (triangAll == 1) {
		sTriangList = freeMem + threadIdx.x * h.leftListSize;
		freeMem  = freeMem + blockDim.x * h.leftListSize;
	}

	//printf ("used sharedMem2=%ld\n", (long)freeMem-(long)sharedMem);


	//printf ("SolveMatrixG: Ok01\n");

	for (int k=0; k < nComp; k++ )
		vmListLocal[POS(k)] = h.vmList[k];

	for (int k=0; k < h.compListSize; k++ ) {
		nGate[POS(k)] = h.n[k];
		hGate[POS(k)] = h.h[k];
		mGate[POS(k)] = h.m[k];
	}

	//__syncthreads();

	//	printf("%f|%f|%f\n", h.tau[0], h.gmax[0], h.esyn[0]);
	//	assert(false);

	ftype dt = h.dt;
	int currStep = h.currStep;
	ucomp nGeneratedSpikes = 0;

	for(int gStep = 0; gStep < nSteps; gStep++ ) {

		//printf ("SolveMatrixG: Ok1\n");
		if (triangAll == 0) {
			updateRhsG(hList, sMulList, sMulListComp, sMulListDest,
					rhsLocal, vmListLocal, nGate, hGate, mGate, sTau, sGmax, sEsyn, freeMem); // RYC
			backSubstituteG(hList, sLeftList, sLeftListLine, sLeftListColumn, rhsLocal, vmListLocal, freeMem); // RYC
		}
		else {

			upperTriangularizeAll(hList, sTriangList, sLeftList, sLeftListLine, sLeftListColumn,
					sLeftStartPos, rhsLocal, vmListLocal, nGate, hGate, mGate, sTau, sGmax, sEsyn,freeMem);
			backSubstituteG(hList, sTriangList, sLeftListLine, sLeftListColumn, rhsLocal, vmListLocal, freeMem); // RYC
		}
		//printf ("SolveMatrixG: Ok2\n");

		for (int k=0; k<nComp; k++) {
			int index = k * nSteps + gStep;
			h.vmTimeSerie[index] = vmListLocal[POS(k)]; // RYC
		}

		currStep = currStep + 1;


		if (vmListLocal[POS(nComp-1)] >= h.threshold && ((currStep * dt) - h.lastSpike) > h.minSpikeInterval) {

			h.spikeTimes[nGeneratedSpikes] = currStep * dt;
			h.lastSpike = currStep * dt;
			nGeneratedSpikes++;
		}

		h.currStep = currStep;
	}

	//printf("spikes=%d neuron=%d\n", nGeneratedSpikes, neuron);
	h.nGeneratedSpikes[neuron] = nGeneratedSpikes;

	//__syncthreads();

	for (int k=0; k<nComp; k++) {
		h.rhsM[k] = rhsLocal[POS(k)];
		h.vmList[k] = vmListLocal[POS(k)]; // RYC
	}

	for (int k=0; k < h.compListSize; k++ ) {
		h.n[k] = nGate[POS(k)];
		h.h[k] = hGate[POS(k)];
		h.m[k] = mGate[POS(k)];
	}

	vmListGlobal[neuron] = vmListLocal[POS(nComp-1)];

	//h.currStep = h.currStep + nSteps;

}
//} // extern C
